package de.intevation.flys.exports;

import java.awt.Color;

import org.apache.log4j.Logger;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.chart.title.TextTitle;
import org.jfree.data.Range;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;

import org.w3c.dom.Document;

import de.intevation.artifacts.Artifact;

import de.intevation.artifactdatabase.state.Facet;

import de.intevation.flys.artifacts.FLYSArtifact;

import de.intevation.flys.artifacts.model.FacetTypes;
import de.intevation.flys.artifacts.model.WQKms;


/**
 * An OutGenerator that generates discharge curves.
 *
 * @author <a href="mailto:ingo.weinzierl@intevation.de">Ingo Weinzierl</a>
 */
public class LongitudinalSectionGenerator
extends      XYChartGenerator
implements   FacetTypes
{
    /** The logger that is used in this generator.*/
    private static Logger logger =
        Logger.getLogger(LongitudinalSectionGenerator.class);

    public static final String I18N_CHART_TITLE =
        "chart.longitudinal.section.title";

    public static final String I18N_CHART_SUBTITLE =
        "chart.longitudinal.section.subtitle";

    public static final String I18N_XAXIS_LABEL =
        "chart.longitudinal.section.xaxis.label";

    public static final String I18N_YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.label";

    public static final String I18N_2YAXIS_LABEL =
        "chart.longitudinal.section.yaxis.second.label";

    public static final String I18N_CHART_TITLE_DEFAULT  = "W-L\u00e4ngsschnitt";
    public static final String I18N_XAXIS_LABEL_DEFAULT  = "km";
    public static final String I18N_YAXIS_LABEL_DEFAULT  = "W [NN + m]";
    public static final String I18N_2YAXIS_LABEL_DEFAULT = "Q [m\u00b3/s]";

    /** The storage for the W series to be drawn in this chart.*/
    protected XYSeriesCollection w;

    /** The storage for the Q series to be drawn in this chart.*/
    protected XYSeriesCollection q;

    protected boolean inverted;


    public LongitudinalSectionGenerator() {
        super();

        this.w = new XYSeriesCollection();
        this.q = new XYSeriesCollection();
    }


    protected String getChartTitle() {
        return msg(I18N_CHART_TITLE, I18N_CHART_TITLE_DEFAULT);
    }

    public boolean isInverted() {
        return inverted;
    }

    public void setInverted(boolean inverted) {
        this.inverted = inverted;
    }

    @Override
    protected void addSubtitles(JFreeChart chart) {
        double[] dist  = getRange();

        Object[] args = new Object[] {
            getRiverName(),
            dist[0],
            dist[1]
        };

        String subtitle = msg(I18N_CHART_SUBTITLE, "", args);
        chart.addSubtitle(new TextTitle(subtitle));
    }


    protected String getXAxisLabel() {
        return msg(I18N_XAXIS_LABEL, I18N_XAXIS_LABEL_DEFAULT);
    }


    protected String getYAxisLabel() {
        return msg(I18N_YAXIS_LABEL, I18N_YAXIS_LABEL_DEFAULT);
    }


    @Override
    protected void addDatasets(JFreeChart chart) {
        XYPlot plot = (XYPlot) chart.getPlot();

        plot.setDataset(0, w);
        plot.setDataset(1, q);
    }


    protected void adjustPlot(XYPlot plot) {
        super.adjustPlot(plot);

        // TODO REMOVE THIS CODE, IF WE HAVE INTRODUCED THEMES!
        XYLineAndShapeRenderer rw = (XYLineAndShapeRenderer)
            plot.getRendererForDataset(w);

        XYLineAndShapeRenderer rq = null;
        try {
            rq = (XYLineAndShapeRenderer) rw.clone();
        }
        catch (Exception e) {
            logger.error(e, e);
        }

        int wNum = w.getSeriesCount();
        int qNum = q.getSeriesCount();

        for (int i = 0; i < wNum; i++) {
            rw.setSeriesPaint(i, Color.BLUE);
        }

        for (int i = 0; i < qNum; i++) {
            rq.setSeriesPaint(i, Color.GREEN);
        }

        plot.setRenderer(0, rw);
        plot.setRenderer(1, rq);
    }


    protected void adjustAxes(XYPlot plot) {
        super.adjustAxes(plot);

        NumberAxis qAxis = new NumberAxis(
            msg(I18N_2YAXIS_LABEL, I18N_2YAXIS_LABEL_DEFAULT));

        plot.setRangeAxis(1, qAxis);
        plot.mapDatasetToRangeAxis(1, 1);

        invertXAxis(plot.getDomainAxis());
    }


    /**
     * This method overrides the XYChartGenerators zoomY method to include the 0
     * value on the Q axis.
     */
    @Override
    protected boolean zoomY(XYPlot plot, ValueAxis axis, Range range, Range x) {
        if (plot.getRangeAxisIndex(axis) == 1) {
            // we want the Q axis to start at 0 if no zooming has been done
            range = new Range(0d, range.getUpperBound());
        }

        return super.zoomY(plot, axis, range, x);
    }


    /**
     * This method inverts the x-axis based on the kilometer information of the
     * selected river. If the head of the river is at kilometer 0, the axis is
     * not inverted, otherwise it is.
     *
     * @param xaxis The domain axis.
     */
    protected void invertXAxis(ValueAxis xaxis) {

        if (inverted) {
            logger.debug("Invert X-Axis.");
            xaxis.setInverted(true);
        }
    }


    public void doOut(Artifact artifact, Facet facet, Document attr) {
        String name = facet.getName();

        logger.debug("LongitudinalSectionGenerator.doOut: " + name);

        if (name == null) {
            logger.error("No facet name for doOut(). No output generated!");
            return;
        }

        FLYSArtifact flys = (FLYSArtifact) artifact;
        Facet        f    = flys.getNativeFacet(facet);

        if (f == null) {
            return;
        }

        if (name.equals(LONGITUDINAL_W)) {
            doWOut((WQKms) f.getData(artifact, context));
        }
        else if (name.equals(LONGITUDINAL_Q)) {
            doQOut((WQKms) f.getData(artifact, context));
        }
        else {
            logger.warn("Unknown facet name: " + name);
            return;
        }
    }


    /**
     * Process the output for W facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     */
    protected void doWOut(WQKms wqkms) {
        logger.debug("LongitudinalSectionGenerator.doWOut");

        XYSeries series = new XYSeries(getSeriesName(wqkms, "W"));

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKms(0));
                logger.debug("End   km: " + wqkms.getKms(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKms(i), wqkms.getW(i));
        }

        w.addSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    /**
     * Process the output for Q facets in a longitudinal section curve.
     *
     * @param wqkms An array of WQKms values.
     */
    protected void doQOut(WQKms wqkms) {
        logger.debug("LongitudinalSectionGenerator.doQOut");

        XYSeries series = new XYSeries(getSeriesName(wqkms, "Q"));

        int size = wqkms.size();

        if (logger.isDebugEnabled()) {
            if (wqkms.size() > 0) {
                logger.debug("Generate series: " + series.getKey());
                logger.debug("Start km: " + wqkms.getKms(0));
                logger.debug("End   km: " + wqkms.getKms(size-1));
                logger.debug("Values  : " + size);
            }
        }

        for (int i = 0; i < size; i++) {
            series.add(wqkms.getKms(i), wqkms.getQ(i));
        }

        q.addSeries(series);

        if (wqkms.guessWaterIncreasing()) {
            setInverted(true);
        }
    }


    protected String getSeriesName(WQKms wqkms, String mode) {
        String name   = wqkms.getName();
        String prefix = name != null && name.indexOf(mode) >= 0 ? null : mode;

        return prefix != null && prefix.length() > 0
            ? prefix + "(" + name +")"
            : name;
    }
}
// vim:set ts=4 sw=4 si et sta sts=4 fenc=utf8 :
